package org.w3c.rdfpic.core;
import org.xml.sax.SAXException;
import org.w3c.rdf.implementation.model.ModelImpl;
import org.w3c.rdf.model.*;
import org.w3c.rdf.util.RDFFactoryImpl;
import org.w3c.rdf.util.ModelConsumer;
import org.xml.sax.InputSource;
import org.w3c.rdf.implementation.model.StatementImpl;
import java.util.LinkedList;
import java.util.Enumeration;
import java.util.Vector;
import java.io.*;
import java.net.URL; 
import java.net.UnknownServiceException;
import java.net.URLConnection;
import java.util.ListIterator;
import java.util.NoSuchElementException;

/**
 * URLVector is essentially a wrapper around Vector so typecasting
 * isn't necessary.  It doesn't offer any new functions.
 */
class URLVector {
    Vector urlsV;
    URLVector(URL[] urls)
    {
	urlsV= new Vector();
	if (urls!=null)
	    {
		urlsV.ensureCapacity(urls.length);
		for(int i=0;i<urls.length;++i)
		    urlsV.add(urls[i]);
	    }
    }
    int size()
    {
	return urlsV.size();
    }
    URL elementAt(int index)
    {
	return (URL)urlsV.elementAt(index);
    }
    void add(URL url)
    {
	urlsV.add(url);
    }
    void remove(int index)
    {
	urlsV.remove(index);
    }
}
/**
 * RdfPicCore provides a unifying shell around the various IO issues,
 * schemas, and the like, while attempting not to implement any UI
 * specific features.  Ideally, it should only take a replacement of Main
 * and an equivalent of SimpleUI to make a utility that could work off the
 * command line, for instance. The core maintains a list of files, and allows
 * the UI to select any given file.  Loading automatically occurs when a
 * file is selected.  All rdf data is sent to an RdfDataProvider, which is
 * also responsible for providing rdf data when saving.  A number of other
 * functions provide the capability to import/extract the raw comment from
 * the image file, or save/load the current rdf data to a text file.
 * <p>
 * The core also maintains a list of Schemas.  Each schema contains information
 * about its name and URI, and a list of predicates and information about those.
 * this means that the rdf data generated is very simple - no complex tree
 * structures, loops, bags or like are produced.  Information about the schema's
 * is revealed through certain functions. Schema's, like files, can be dynamically
 * loaded at runtime.
 *
 */

public class RdfPicCore
{
    /**
     * contains an explanation of the current error, or null if there isn't any.
     */
    String errMsg;
    /**
     * all loaded schemas.
     */
    RdfSchema[] schemas;
    /**
     * an object responsible for parsing and serialzing a schema.
     */
    SchemaFileDrv schemadrv;
    /**
     * stores whether a schema is active or not. Inactive schemas are not saved into the output.
     */
    boolean[] schemaEnabled;
    /**
     * an object responsible for keeping track of the rdf data.
     */
    RdfDataProvider dataSource;
    /**
     * Lists of objects to notify when an event occurs.
     */
    LinkedList schemaListeners,imgListeners;
    /**
     * the list of files.
     */
    URLVector picV;
    /**
     * The handler is responsible for embedding text data into an image.
     */
    ImgMetaHandler handler;
    /**
     * an object that deals with placing the loading activity into another thread.
     */
    ImageLoader loadThread;
    /**
     * the index of the active file (in picV).
     */
    int index;
    /**
     * stores an rdf-model of the statements that weren't understood.  They are inserted into the output unchanged.
     */
    Model extras;
    /**
     * The code of the languge used in the rdf Description.
     */
    String langcode;
    private void die(Throwable exp,String msg)
    {
	if(msg!=null)
	    System.err.print(msg);
	if(exp!=null)
	    exp.printStackTrace();
	System.exit(1);
    }
    /**
     * Initializes the core with the given schemas, pictures, and schema driver.
     */
    public RdfPicCore(SchemaFileDrv schemadrv, URL[] schemaURLs, URL[] picURLs) throws SetupIncompleteException, IOException
    {
	//we need an object to load SchemaFiles
	if (schemadrv==null) throw new SetupIncompleteException();
	if (schemaURLs==null) schemaURLs=new URL[0];
	picV=new URLVector(picURLs);
	langcode="en";

	this.schemadrv=schemadrv;

	//The handler for doing meta data stuff.
	handler = new ImgMetaHandler();
	index=-1;
	
	//Load all Schema Definitions
	schemas = new RdfSchema[schemaURLs.length];
	schemaEnabled = new boolean[schemaURLs.length];
	for(int i=0;i<schemaURLs.length;i++)
	    {
		schemaEnabled[i]=true;
		schemadrv.setURL(schemaURLs[i]);
		schemas[i]=schemadrv.read();
	    }
	
	schemaListeners=new LinkedList();
	imgListeners   =new LinkedList();

	loadThread = new ImageLoader(this);
	loadThread.start();
	try {
	    loadThread.setPriority(Thread.MIN_PRIORITY+1);
	} catch(Exception e) {} //if it doesn't work, only means UI will be less snappy
	extras=null;
    }
    
    public void setRdfDataProvider(RdfDataProvider rdp){dataSource=rdp;}    
    public void addSchemaListener(SchemaListener sl)   {schemaListeners.add(sl);}
    public void addImgListener(ImgListener il)         {imgListeners.add(il);}
    /**
     * Returns whether or not the schema with the given index is enabled.
     */
    public boolean isEnabled(int schemaIndex)          {return schemaEnabled[schemaIndex];}    
    public int numSchemas()                            {return schemas.length;}
    public String schemaName(int i)                    {return schemas[i].name;}
    public int numPredicates(int i)                    {return schemas[i].predicates.length;}
    public String predicateName(int sI, int pI)        {return schemas[sI].predicates[pI];}
    public String defaultVal(int sI, int pI)           {return schemas[sI].defaults[pI];}
    /**
     * Sometimes, you want special functionality in a schema. A schema can be marked "builtin"
     * which has no impact on the rdf, but allows special decisions to be made on its basis.
     */
    public boolean builtinSchema(int sI)               {return schemas[sI].builtin!=0;}
    /**
     * Whether the predicate has to be treated specially.
     */
    public boolean isBuiltinPredicate(int sI, int pI)    {return schemas[sI].isBuiltin[pI];}
    /**
     * An Explanation of the purpose of the predicate.
     */
    public String predExpl(int sI, int pI)             {return schemas[sI].expl[pI];}
    /**
     * Returns the currently active file's  index.
     */
    public int getPos()                                {return index;}
    public int numFiles()                              {return picV.size();}
    /**
     * gets the language (of the rdf) of the active file.
     */
    public String getLang()                            {return langcode;}
    /**
     * sets the language (of the rdf) of the active file.
     */
    public void setLang(String lang)                   {langcode=lang;}
    /**
     * Returns the filename of the <i>i</i>th file in the file list, or "<none>"
     * if the index is out of range.
     */
    public String fileName(int i)                      
    {
	if (i>=0 && i<picV.size()) return picV.elementAt(i).toString();
	else return "<none>";
    }
    /**
     * Go to the file i further than the current file.  For example, <code>jumpTo(-1)</code> would go to the
     * previous file.
     */
    public void jumpRel(int i)                         {jumpTo(index+i);}
    /**
     * Returns the URL of the currently active file.
     */
    public URL getURL()                                
    {
	if (index>=0 && index<picV.size())
	    return picV.elementAt(index);
	else return null;
    }

    /**
     * Adds a schema to the schema list's end.  Schema's cannot be removed again, but disabled, if necessary.
     */
    public void addSchema(URL schemaURL) throws IOException
    {  
	RdfSchema schema;
	schemadrv.setURL(schemaURL);
	schema=schemadrv.read();
	if(schema==null) throw new IOException("Syntax or IO error reading schema file.");
	RdfSchema[] newschemas;
	boolean[] newEnabled;
	newschemas=new RdfSchema[schemas.length +1];
	newEnabled=new boolean[schemas.length+1];
	System.arraycopy(schemas,0,newschemas,0,schemas.length);
	System.arraycopy(schemaEnabled,0,newEnabled,0,schemas.length);
	newschemas[schemas.length]=schema;
	newEnabled[schemas.length]=true;
	schemas=newschemas;
	schemaEnabled=newEnabled;
	
	ListIterator li = schemaListeners.listIterator(0);
	while (li.hasNext())
	    {
		try {
		    ((SchemaListener)li.next()).addingSchema();
		} catch(NoSuchElementException e) {}//hasNext implies next
	    }
    }

    /**
     * Enables a schema, meaning the rdf-info contained in it will be output in a save
     * operation.
     */
    public void enableSchema(int i)
    {
	if(i<0 || i >= schemas.length) return;
	if (schemaEnabled[i]) return;
	schemaEnabled[i]=true;
	ListIterator li = schemaListeners.listIterator(0);
	while (li.hasNext())
	    {
		try {
		    ((SchemaListener)li.next()).enablingSchema(i);
		} catch(NoSuchElementException e) {}//hasNext implies next works.
	    }
    }
    /**
     * Disables a schema.  That schema's contents will not be included in any output.
     */
    public void disableSchema(int i)
    {
	if(i<0 || i >= schemas.length) return;
	if (!schemaEnabled[i]) return;
	schemaEnabled[i]=false;
	ListIterator li = schemaListeners.listIterator(0);
	while (li.hasNext())
	    {
		try {
		    ((SchemaListener)li.next()).disablingSchema(i);
		} catch(NoSuchElementException e) {}//hasNext implies next works
	    }
    }
    /**
     * Exports the RDF to a text file.
     */
    public void exportRDF(File file) throws IOException
    {
	Writer writer;
	writer = new BufferedWriter(new FileWriter(file));
	try {
	    (new RDFSerial()).serialize(attachLangInfo(getData("")),writer);
	} catch (ModelException me) {
	    die(me,"?Unfinished: Exception not checked");//FIXME
	}
    }
    /**
     * Imports RDF from a text file.
     */
    public void importRDF(File file) throws RdfPicError,IOException
    {
	extras=null;
	Reader reader;
	reader= new BufferedReader(new FileReader(file));
	loadData(reader);
    }
    /**
     * Returns whether the file is writable or not - falsely returns true when in doubt (esp. over HTTP).
     */
    public boolean canWrite()
    {
	return handler.canWrite();
    }
    /**
     * Extracts the raw contents of the image's comment into the given file.
     */
    public void extractRaw(File file) throws RdfPicError/*When image isn't in legal state*/,IOException/*file is evil*/
    {
	if (!handler.canRead()) throw new RdfPicError("Cannot read from image");
	Writer writer;
	writer = new BufferedWriter(new FileWriter(file));
	writer.write(handler.read());
	writer.close();
    }
    /**
     * Inserts the unaltered contents of the given file into the image's comment.
     */
    public void insertRaw(File file) throws RdfPicError,IOException
    {
	if (!handler.canWrite()) throw new RdfPicError("Cannot write to image");
        Reader reader;
	reader = new BufferedReader(new FileReader(file));
	StringBuffer strbuf=new StringBuffer();
	char[] charbuf=new char[1024];
	int count;
	while ((count=reader.read(charbuf))!=-1)
	    strbuf.append(charbuf,0,count);
	handler.write(strbuf.toString());
	reader.close();
    }
    /**
     * Adds a URL of a picture to the end of the file list.
     */
    public void addURL(URL pic)
    {
	int newindex;
	if(pic==null) return;
	newindex=picV.size();
	picV.add(pic);
	ListIterator li = imgListeners.listIterator(0);
	while (li.hasNext())
	    {
		try {
		    ((ImgListener)li.next()).imgVUpdate();
		} catch(NoSuchElementException e) {}//li.next cannot fail as we check hasNext every iteration
	    }
    }
    /**
     * Removes the URL of the picture at the given index from the file list.
     * Has no effect if the index is out of range, or the index of the active file.
     */
    public void removeURL(int i)
    {
	if(i < 0 || i >= picV.size()||index == i) return;
	picV.remove(i);
	if (index >i) index--;
	ListIterator li = imgListeners.listIterator(0);
	while (li.hasNext())
	    {
		try {
		    ((ImgListener)li.next()).imgVUpdate();
		} catch(NoSuchElementException e) {}//li.next cannot fail as we check hasNext every iteration
	    }
    }
    /**
     * Selects and makes active a file with the given index.
     */
    public void jumpTo(int i)
    {
	if(i==index) return;
	if(i<0 || i>=picV.size()) i=-1;
	errMsg="Still loading...";
	ListIterator li = imgListeners.listIterator(0);
	while (li.hasNext())
	    {
		try {
		    ((ImgListener)li.next()).isJumping(i);
		} catch(NoSuchElementException e) {}//li.next cannot fail as we check hasNext every iteration
	    }
	index=i;
	loadThread.reload();//simply calls finishJump but in a different thread.
    }
    /**
     * Returns the image of the current file.
     * @return the image (java.awt.Image), or <code>null</code> if none loaded.
     * @exception RdfPicError when called while an image is being loaded.
     */
    public java.awt.Image getImg() throws RdfPicError
    {
	if(errMsg!=null) throw new RdfPicError("No Image loaded\n"+errMsg);
	return handler.getImg();
    }

    /**
     * Retrieves the size of the current file.
     * @return the size in bytes
     * @exception RdfPicError when called while an image is being loaded.
     * @exception PHException when the current image couldn't be successfully read (for instance when none is loaded).
     */
    public int getImgFileSize() throws PHException, RdfPicError
    {
	if(errMsg!=null) throw new RdfPicError("No Image loaded\n"+errMsg);
	return handler.getFileSize();
    }

    /**
     * Retrieves the width of the current image.
     * @return the width in pixels, -1 if unknown
     * @exception RdfPicError when called while an image is being loaded.
     * @exception PHException when the current image couldn't be successfully read (for instance when none is loaded).
     */
    public int getImgWidth() throws PHException, RdfPicError
    {
	if(errMsg!=null) throw new RdfPicError("No Image loaded\n"+errMsg);
	return handler.getImg().getWidth(null);
    }

    /**
     * Retrieves the height of the current image.
     * @return the height in pixels, -1 if unknown
     * @exception RdfPicError when called while an image is being loaded.
     * @exception PHException when the current image couldn't be successfully read (for instance when none is loaded).
     */
    public int getImgHeight() throws PHException, RdfPicError
    {
	if(errMsg!=null) throw new RdfPicError("No Image loaded\n"+errMsg);
	return handler.getImg().getHeight(null);
    }

    void finishJump()//In other Thread!!!
    {
	extras=null;
	try {
	    if (index!=-1) handler.setURL(picV.elementAt(index));//This loads the new URL and the data it contains.
	    else  handler.setURL(null);
	    errMsg=null;
	} 
	catch (PH_noRead pnr) {errMsg="Unable to open for reading\n("+pnr.getMessage()+")";}
	catch (PH_IOError pioe) {errMsg="I/O error\n("+pioe.getMessage()+")";}
	catch (PH_noURL pnurl) {errMsg="No URL specified.";}
	catch (PH_unknownProtocol pup) {errMsg="Unknown Protocol\n("+pup.getMessage()+")";}
	catch (PH_badURL pbu) {errMsg="Bad URL\n("+pbu.getMessage()+")";}
	catch (PH_notFound pnf) {errMsg="File not found.";}
	catch (PHWarning phw) {errMsg=null;}
	catch (PHException phe) {errMsg=phe.getMessage();}
	catch (UnknownImgTypeException uite) {errMsg="Can't handle file: Unknown Image Type";}

	ListIterator li = imgListeners.listIterator(0);
	while (li.hasNext())
	    {
		try {
		    ((ImgListener)li.next()).hasJumped(errMsg);
		} catch(NoSuchElementException e) {
		    die(e,"?Impossible: next cannot fail after hasNext==true");
		}//li.next cannot fail as we check hasNext every iteration
	    }
    } 
    /**
     * Saves the rdf currently available to the image.
     */
    public void saveData() throws RdfPicError,IOException
    {
	if(errMsg!=null) throw new RdfPicError("You need to have an image correctly loaded first\n("+errMsg+")");
	if(!handler.canWrite()) {
	    if(index!=-1)
		throw new RdfPicError("Can't write to image (no write permissions?)");
	    else
		throw new RdfPicError("Can't write to image (no image loaded)");
	}
	try {
	Rdf2ImgMeta.save(attachLangInfo(getData("")),handler,new RDFSerial());
	} catch(ModelException me) {
	    die(me,"?Impossible: problem in self-generated model");
	} catch(org.w3c.rdf.syntax.SerializationException se) {
	    die(se,"?Internal error.");
	}
    }
    /**
     * Loads the rdf embedded in the image.
     */
    public void loadData() throws RdfPicError,IOException
    {
	if(handler.getURL()==null) return;
	if(!handler.canRead()) throw new RdfPicError("Unable to load comment from image");
	Reader reader;
	String str = handler.read();
	if (str==null) str="";
	reader = new StringReader(str);
	loadData(reader);
    }

    private void loadData(Reader reader) throws RdfPicError
    {
	Enumeration e;
	InputSource source;
	RDFFactoryImpl f = new RDFFactoryImpl();
	Model model = f.createModel();
	extras= f.createModel();
	source = new InputSource(reader);
	//	source.setSystemId(picV.elementAt(index).toString());
	try {
	    f.createParser().parse(source,new ModelConsumer(model));
	} catch (SAXException saxe) {
	    throw new RdfPicSyntaxException("Syntax Error - Comment does not contain RDF?");
	} catch (ModelException me) {
	    die(me,"?Impossible: ModelImpl won't throw in ModelConsumer");
	}
	e=null;
	try {
	    e = model.elements();
	} catch (ModelException err) {
	    die(err,"?Impossible: ModelImpl.elements does throw");
	}

	while (e.hasMoreElements())
	    {
		Statement temp=null;
		Resource subject=null,predicate=null;
		RDFNode object=null;

		try {
		    temp=(Statement)e.nextElement();
		} catch(Exception err) {
		    die(err,"?Impossible: e.nextElement works because e.hasMoreElements()==true");
		}
		try {
		    subject=temp.subject();
		    predicate=temp.predicate();
		    object=temp.object();
		} catch(Exception err) {
		    die(err,"?Impossible: StatementImpl does not throw");
		}
		
		try {
		    if (subject.getURI().trim().equals("") || subject.getURI().trim().equals("."))
			if(predicate.getURI().equals("http://sophia.inria.fr/~enerbonn/rdfpiclang#xmllang")) {
			    langcode=object.getLabel();
			    continue;
			}
		} catch(ModelException me) {
		    die(me,"?Unlikely: Unable to extract strings from statement");
		}
		       
		String predURI=null,predLabel=null,predTotal=null;
		int cutpos=-1;

		try {
		    predTotal=predicate.getURI();
		    for(int i=0;i<predTotal.length();++i)
			if(predTotal.charAt(i)=='#')
			    {
				cutpos=i;
				break;
			    }
		    if(cutpos==-1)
			{
			    extras.add(temp);
			    continue;
			}
		    predURI=predTotal.substring(0,cutpos);
		    predLabel = predTotal.substring(cutpos+1);
		} catch (Exception err) {
		    die(err,"?Impossible: ResourceImpl does not throw");
		}

		boolean OK=false;
		for(int schemaI=0;schemaI<schemas.length;schemaI++)
		    if(!schemas[schemaI].URI.equals(predURI)) continue;
		    else 
			for(int predI=0;predI<schemas[schemaI].predicates.length;predI++)
			    if(!schemas[schemaI].predicates[predI].equals(predLabel)) continue;
			    else
				{
				    OK=true;
				    try {
					dataSource.setVal(schemaI,predI,object.getLabel());
				    } catch (ModelException me) {
					die(me,"?Impossible: RDFNodeImpl does not throw");
				    }
				}
		try {
		    if(!OK)	extras.add(temp);
		} catch(ModelException me) {
		    die(me,"?Unlikely: ModelImpl using StatementImpl should never throw here");
		}
	    }
    }
    /**
     * Returns a Model of the data according to RdfDataProducer.
     */
    Model getData(String subject) throws ModelException
    {
	ModelImpl model = new ModelImpl();
	String val;
	int count=0;
	for(int schemaI=0;schemaI<schemas.length;schemaI++) {
	    if(!schemaEnabled[schemaI]) continue;
	    RdfSchema schema = schemas[schemaI];
	    for(int propI=0;propI<schemas[schemaI].predicates.length;propI++)
		if((val=dataSource.valOf(schemaI,propI))!=null) if(val.length()>0)
		    addTo(model,subject,schema.URI + "#" + schema.predicates[propI],val);
	}
	if (extras==null) return model;
	Enumeration e = extras.elements();
	while(e.hasMoreElements()) model.add((Statement)e.nextElement());
	return model;
    }
    private static void addTo(Model model, String sub, String pred, String val) throws ModelException {
	NodeFactory nodeFactory = model.getNodeFactory();
	model.add(new StatementImpl(nodeFactory.createResource(sub), nodeFactory.createResource(pred), nodeFactory.createLiteral(val)));
    }

    Model attachLangInfo(Model model) throws ModelException {
	addTo(model,"","http://sophia.inria.fr/~enerbonn/rdfpiclang#xmllang",langcode);
	return model;
    }

}
